#! /bin/sh
#
# netfilter-init Set up netfilter packet filtering for Debian Systems
#
# 2001-02-22 Marc Haber            Initial Version

# this script has one action more than a normal init script. the 
# action start-pre closes the packet filter completely. This is
# intended to prepare the system for bringing up the network interfaces
# and is being used by another init script that is invoked before network
# initialization.

PATH=/sbin:/usr/sbin:/bin:/usr/bin
SCRIPT=/usr/sbin/netfilter-init
DESC="netfilter packet filtering"

test -f $SCRIPT || exit 0

set -e

case "$1" in
  start|reload|force-reload|restart)
#	if [ -e /lib/modules/`uname --release`/ipv4/iptables.o ]; then
#	  modprobe iptables
#	fi
  	echo -n "Building $DESC: "
	$SCRIPT up
	echo "done."
	;;
  start-pre)
	echo -n "Preparing $DESC for initialization - closing all filters: "
	$SCRIPT --yes close-completely
	echo "done."
	;;
  stop)
	echo -n "Closing $DESC: "
	$SCRIPT close
	echo "closed. (ssh access has been left open)"
	;;
  *)
	N=/etc/init.d/$NAME
	echo "Usage: $N {start|stop|restart|reload|force-reload}" >&2
	exit 1
	;;
esac

exit 0
