/*
 * Copyright (c) 1999 Marc Haber <Marc.Haber@gmx.de>
 * All rights reserved.
 *
 * This software's redistribution and use in source and binary forms
 * is defined in the GNU General Public License.
 */

//#define DEBUG 1
#undef DEBUG

#include <iostream.h>
#include <regex.h>
#include <string>
#include <vector>
#include <algo.h>
#include <stdlib.h>

int DoChainExpansion = 0;

string ConcatStringArray(char **arr, int num)
{
		int i;
		string rval="";
		
		for( i=0; i<num; i++ )
		  {
				  rval += arr[i];
                  rval += " ";
		  }
		
		return rval;
}

string GetEnv(const string name)
{
  char *pEnv = getenv(name.c_str());
  if( pEnv == NULL )
  {
	return "";
  }
  else
  {
	return pEnv;
  }
}


string Trim(string s)
{
 string::iterator i=s.begin();
		
  while(isspace(*i++) && i!=s.end());
	
  if( i-- != s.begin() ) { s.erase(s.begin(),i); }
		
  i=s.end();
  while(isspace(*--i) && i!=s.begin());
  if( ++i != s.end() ) { s.erase(i,s.end()); }

  if( s == " " ) s="";
  return s;
}

void ReplaceString( string &In, string Search, string Target)
{
  #ifdef DEBUG
  cerr << "ReplaceString(" << In << "," << Search << "," << Target << ");\n";
   cerr << "In.find(Search)==" << In.find(Search) << ":\n";
   cerr << "Search.length()==" << Search.length() << ":\n";
  #endif

  string::size_type pos=0;
  while( (pos=In.find(Search,pos)) != string::npos )
  {
	In.replace(pos,Search.length(),Target);
	pos += Target.length();
  }
  #ifdef DEBUG
  cerr << "Output: " << In << ":\n";
  #endif
}


string FindInterfaceFromEnvironment( string Str, string envvar )
{		  
  // determine if we have a mex containing an interface
		  
  string Env=GetEnv(envvar)+" ";
  string This;
  Str = Str+" ";
  #ifdef DEBUG
  cerr << "Str:  " << Str  << ":\n";
  #endif
  while( Env!="" )
  {
    int FirstSpace=Env.find_first_of(' ');
    This=Env.substr(0,FirstSpace);
    Env.replace(0,FirstSpace+1,"");
    Env=Trim(Env)+" ";
    This=Trim(This)+" ";
	#ifdef DEBUG
	cerr << "Env:  " << Env  << ":\n";
	cerr << "This: " << This << ":\n";
	#endif
	if( Str.find(This)!=string::npos )
	{
	  This=Trim(This);
	  #ifdef DEBUG
	  cerr << "String " << This << ": found\n";
	  #endif
	  return This;
	}
  }
  return "";
}


void MultiplyExpandSingleString( vector<string> &vec, int i)
{
  #ifdef DEBUG
  cerr << "MultiplyExpandSingleString: vec[" << i << "] == " << vec[i] << "\n";
  #endif

  string source = Trim(vec[i]);
  
  string target = "";
  
  string::size_type idx;
		
  while( source!="" )
  {
  
    #ifdef DEBUG
    cerr << "source: " << source << ":\n";
	cerr << "target: " << target << ":\n";
    #endif
  
    idx = source.find('{');

    string start = Trim(source.substr(0, idx));
    target += start;
	
    #ifdef DEBUG
    cerr << "{ at: " << idx << "\n";
	cerr << "target: " << target << ":\n";
	cerr << "start: " << start << ":\n";
    #endif
	
    if( idx == string::npos )
  	{
	  // no mex inside source
      #ifdef DEBUG
      cerr << "no mex inside source\n";
      #endif
      source="";
	}
	else
	{
	  // we have a mex. remove start and {
	  source.erase(0,idx);
	  
	  // get middle part to mex and remove it from source
	  
	  idx = source.find('}');
	  string mex = Trim(source.substr(1, idx-1));
	  source.erase(0,++idx);

      	  #ifdef DEBUG
	  cerr << "source: " << source << ":\n";
          #endif
      	  source = Trim(source);

      	  #ifdef DEBUG
	  cerr << "mex found\n";
	  cerr << "mex: " << mex << ":\n";
	  cerr << "source: " << source << ":\n";
      	  #endif

	  if( mex != "" )
	  {
	    // mex found
		
		int count=0;
		
		while( mex != ""  && count++<500)
		{
		  idx = mex.find(',');
		  string thismex = Trim(mex.substr(0, idx));
		  string newcommand = target + ' ' + thismex + ' ' + source;
				
                  #ifdef DEBUG
	          cerr << "mex.length(): " << mex.length() << "\n";
	          cerr << "mex: " << mex << "\n";
                  cerr << ", at: " << idx << "\n";
		  cerr << "thismex: " << thismex << ":\n";
		  cerr << "newcommand: " << newcommand << ":\n";
                  #endif

		  if( idx!=string::npos )
		  {
	            mex.erase(0,++idx);
		  }
		  else
		  {
		    mex="";
		  }
			
                  #ifdef DEBUG
	          cerr << "mex.length() after: " << mex.length() << "\n";
                  #endif
		  if( thismex != "" )
		  {
		     if( mex.length()>1 )
		     {
		        // this is the case if another ',' has been found
			// previously. Remove everything including the ',',
			// push what we have as a new target

                        #ifdef DEBUG
			cerr << "make new target " << newcommand << ".\n";
                        #endif
			
			vec.push_back(newcommand);
		      }
		      else
		      {
	                // this is the case if the current mex is the last one.
			// stuff last mex element back to source and start over
			
			source = newcommand;
			vec[i] = source;
			target = "";

			#ifdef DEBUG
			cerr << "stuff back to source: " << source << "\n";
 			#endif

			// set mex to empty and make the loop break
			mex = "";
		      }
		  }
			
		  mex = Trim(mex);
          #ifdef DEBUG
          cerr << "mex now: " << mex << ":\n";
          #endif
		}
	  }
	  else
	  {
	    // empty mex, delete current command
        #ifdef DEBUG
        cerr << "empty mex\n";
		#endif
		source="";
		target="";
		vec[i]="";
	  }
	}
  }
 
#ifdef DEBUG
  cerr << "End MultiplyExpandSingleString\n";
#endif
}

void ExpandPlaceholders( string &In )
{
#ifdef DEBUG
	cerr << "ExpandPlaceholders( " << In << ")\n";
#endif
  string ThisInterface = FindInterfaceFromEnvironment(In,"INTERFACES");

  if( ThisInterface != "" )
  {
	// we have a string containing an interface
	#ifdef DEBUG
	cerr << "Doing Replacements on " << In << ":\n";
	cerr << "ThisInterface: " << ThisInterface << ":\n";
	#endif

	string DynamicIface = FindInterfaceFromEnvironment(In, "DYNAMIC_IFACES");
	
	if( DynamicIface != "" )
	{
	  // we have a dynamic interface. Replace all IP with "0.0.0.0[/0]"
	  // Netmask is always set to /0 since interface specific chains have
	  // their IP source/dest already had verified.
   	  ReplaceString(In, "ipnmofif",   "0.0.0.0/0" );
	  ReplaceString(In, "ipofif",     "0.0.0.0/0" );
	  ReplaceString(In, "targetofif", "0.0.0.0/0" );
	  ReplaceString(In, "bcnmofif",   "0.0.0.0/0" );
	  ReplaceString(In, "bcofif",     "0.0.0.0/0" );
	}
	else
	{
   	  ReplaceString(In, "ipnmofif",   GetEnv("IPNMOF_"  +ThisInterface));
	  ReplaceString(In, "ipofif",     GetEnv("IPOF_"    +ThisInterface));
	  ReplaceString(In, "targetofif", GetEnv("TARGETOF_"+ThisInterface));
	  ReplaceString(In, "bcnmofif",   GetEnv("BCNMOF_"  +ThisInterface));
	  ReplaceString(In, "bcofif",     GetEnv("BCOF_"    +ThisInterface));
	}
	ReplaceString(In, "iface",        GetEnv("NAMEOF_"  +ThisInterface));
	ReplaceString(In, " "+ThisInterface+" ",  " "+GetEnv("NAMEOF_"  +ThisInterface)+" ");
  }
}
  
  
void ExpandChains( string &In )
/* replaces the chain a rule is to be modified in with an interface-
 * specific chain if the interface to a rule is a dynamic interface.
 * This function has knowledge of ipchains syntax and needs to be
 * changed if a different packet filter system is to be used.
 * 
 * if the interface has found to be dynamic, the set number is stripped
 * from the rule name and the interface name is prepended instead. */
{
  if( !DoChainExpansion )
  {
	return;
  }
 
  string chain_opts = "ADCRI"; // ipchains-options that are followed by a chain name
  string iface_opts="i";       // ipchains-options that are followed by an interface name

  regex_t reComp1;
  string regex = ".*[[:space:]]-["+iface_opts+"][[:space:]]+([^[:space:]]*).*";

  #ifdef DEBUG
  cerr << "ExpandChains(" << In << ")\n";
  #endif
  
  regcomp( &reComp1, regex.c_str(), REG_EXTENDED );
  regmatch_t Match[2];
  if( regexec( &reComp1, In.c_str(), 2, Match, 0) == 0)
  {
	string Ifacename=In.substr(Match[1].rm_so,Match[1].rm_eo-Match[1].rm_so);
    #ifdef DEBUG
    cerr << "Ifacename: " << Ifacename << ":\n";
    #endif

    string DynIfaces = GetEnv("DYNAMIC_IFACES");
    if( DynIfaces.find( Ifacename ) != string::npos )
    {
      #ifdef DEBUG
	  cerr << Ifacename << " is dynamic, doing chain expansion.\n";
	  #endif
	
      regex_t reComp2;
	  string regex = ".*[[:space:]]-["+chain_opts+"][[:space:]]+([^[:space:]]*).*";
	  regcomp( &reComp2, regex.c_str(), REG_EXTENDED );
	  if( regexec( &reComp2, In.c_str(), 2, Match, 0 ) == 0 )
	  {
 	    string Chain=In.substr(Match[1].rm_so,Match[1].rm_eo-Match[1].rm_so);
        #ifdef DEBUG
        cerr << "Chain: " << Chain << ":\n";
        #endif

	    string NewChain = Chain+""+Ifacename;
		NewChain.replace(0,2,"");
	  
	    /* insert new chain name into rule
	     * this will break if the chain name appears twice in ipchains_parm
	     * can't think about a situation where this might happen, though. */

        #ifdef DEBUG
        cerr << "replacing " << Chain << " with " << NewChain << ".\n";
	    #endif

	    In.replace(In.find(Chain),Chain.length(),NewChain);
		  
        #ifdef DEBUG
        cerr << "new parameter list: " << In << ":\n";
	    #endif
	  }
	}
  }
#ifdef DEBUG
  else
	  {
		  cerr << "no iface option found\n";
	  }
#endif	   
  
  regfree( &reComp1 );

#ifdef DEBUG
	cerr << "end ExpandChains\n";
#endif
	
}


vector<string> ExpandThisString(string input)
{
  vector<string> rval;
  rval.push_back(input);

  // iterate through all strings and expand them

  for( int i = 0; i < rval.size(); i++)
  {
    #ifdef DEBUG
	cerr << "Enter Loop\n";
	cerr << "Calling MultiplyExpandSingleString for string "
		 << rval[i] << ":\n";
    #endif
    MultiplyExpandSingleString( rval, i );

	// Do Chain Expansion. The function will be a no-op if chain expansion
	// is turned off.
	
	ExpandChains( rval[i] );
	
	// Do Placeholder expansion. This must be done after chain expansion
	// because chain expansion may remove placeholders if the mex applies
	// to a dynamic interface.
	
	ExpandPlaceholders( rval[i] );

	// Placeholder expansion may have yielded additional mex expressions
	// expand them too.
	
	MultiplyExpandSingleString( rval, i );
  }

  #ifdef DEBUG
  cerr << "End loop\n";
  #endif
	  			
  return rval;
}


int main(int argc, char* argv[])
{

  // -c as first argument turns off chain expansion
  
  string Opt = argv[1];
  int first = 1;
  DoChainExpansion=1;
  if( (Opt == "-c") || (Opt == "-C") )
  {
	DoChainExpansion=0;
	first++;
  }
  
  string input = ConcatStringArray(argv+first, argc-first);

  #ifdef DEBUG
  cerr << "IN: " << input << "\n" ;
  #endif

  vector<string> Expanded = ExpandThisString(input);

  #ifdef DEBUG
  cerr << "OUT:\n";
  #endif
	  
  vector<string>::const_iterator Ei;
  string pre="";
  for( Ei = Expanded.begin(); Ei != Expanded.end(); ++Ei)
  {
	if( Trim(*Ei) != "" )
	{
	  cout << pre << *Ei;
	  pre=" ; ";
	}
  }
  
  return 0;
}
